"""
The io module contains various utility methods for loading/saving Python objects/dicts in various formats.
Currently, there is support for the most frequent data storing formats:
    - JSON (+gzip)
    - pickled Python objects (+gzip)
"""

import gzip
import json
import pickle
from pathlib import Path
from typing import Union

import numpy as np
import yaml
from PIL import Image

from elias.util.fs import ensure_file_ending, ensure_directory_exists_for_file

PathType = Union[str, Path]


# =========================================================================
# Zipped JSON (.json.gz)
# =========================================================================

def save_zipped_json(obj: dict, path: PathType, suffix: str = 'json.gz'):
    """
    Parses the given Python dict into json, zips it and stores it at the specified location.
    Per default, the path will have a suffix 'json.gz'.

    Parameters
    ----------
        obj: dict
            The JSON (represented as python dict) to be zipped and stored
        path: str
            Where to store the file
        suffix: str, default 'json.gz'
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with gzip.open(path, 'wb') as f:
        json.dump(obj, f)


def load_zipped_json(path: PathType, suffix: str = 'json.gz') -> dict:
    """
    Loads the specified compressed JSON file from `path` and extracts it.
    Per default, the path name is assumed to have a suffix 'json.gz'.

    Parameters
    ----------
        path: str
            Path to the zipped JSON to load
        suffix: str, default 'json.gz'
            File name suffix

    Returns
    -------
        The contents of the zipped JSON as a Python dict
    """

    path = ensure_file_ending(path, suffix)
    with gzip.open(path, 'rb') as f:
        return json.load(f)


# =========================================================================
# Zipped pickles (.p.gz)
# =========================================================================

def save_zipped_object(obj: object, path: PathType, suffix: str = 'p.gz'):
    """
    Pickles, zips and stores an arbitrary python object at the specified `path`.
    Per default, the file will have a suffix 'json.gz'.

    Parameters
    ----------
        obj: object
            The Python object to be zipped and stored
        path:
            Where to store the file
        suffix: str, default 'p.gz'
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with gzip.open(path, 'wb') as f:
        pickle.dump(obj, f)


def load_zipped_object(path: PathType, suffix: str = 'p.gz') -> object:
    """
    Loads zipped and pickled Python object stored at `path`.
    Per default, the file name is assumed to have a suffix 'p.gz'.

    Parameters
    ----------
        path: str
            Path to the zipped Python object that should be loaded
        suffix: str, default 'p.gz'
            File name suffix

    Returns
    -------
        The un-pickled Python object
    """

    path = ensure_file_ending(path, suffix)
    with gzip.open(path, 'rb') as f:
        return pickle.load(f)


# =========================================================================
# Pickled objects (.p)
# =========================================================================

def save_pickled(obj: object, path: PathType, suffix: str = 'p'):
    """
    Pickles and stores an arbitrary python object at the specified `path`.
    Per default, the file will have a suffix 'p'.

    Parameters
    ----------
        obj: object
            The Python object to be pickled and stored
        path: str
            Where to store the file
        suffix: str, default 'p'
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with open(f"{path}", 'wb') as f:
        pickle.dump(obj, f)


def load_pickled(path: PathType, suffix: str = 'p') -> object:
    """
    Loads a pickled Python object stored at the specified `path`.
    Per default, the file name is assumed to have a suffix 'p'.

    Parameters
    ----------
        path: str
            Path to the pickled Python object that should be loaded
        suffix: str, default 'p'
            File name suffix

    Returns
    -------
        The un-pickled Python object
    """

    path = ensure_file_ending(path, suffix)
    with open(path, 'rb') as f:
        return pickle.load(f)


# =========================================================================
# JSON (.json)
# =========================================================================

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def save_json(obj: dict, path: PathType, suffix: str = 'json'):
    """
    Stores the given Python dict in JSON format at `path`.
    Per default, the file will have a suffix 'json'.

    Parameters
    ----------
        obj: dict
            The Python dict to store
        path: str
            Where to store the file
        suffix:
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with open(path, 'w') as f:
        json.dump(obj, f, indent=4, cls=NumpyEncoder)


def load_json(path: PathType, suffix: str = 'json') -> dict:
    """
    Loads and parses the given JSON file and returns it as a Python dict.
    Per default, the file name is assumed to have a suffix 'json'.

    Parameters
    ----------
        path: str
            Path to the JSON file
        suffix: str, default 'json'
            File name suffix

    Returns
    -------
        Contents of the JSON file as Python dict
    """

    path = ensure_file_ending(path, suffix)
    with open(path, 'r') as f:
        return json.load(f)


# =========================================================================
# YAML (.yaml)
# =========================================================================

def save_yaml(obj: dict, path: PathType, suffix: str = 'yaml'):
    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with open(path, 'w') as f:
        yaml.dump(obj, f)


def load_yaml(path: PathType, suffix: str = 'yaml') -> dict:
    path_with_suffix = ensure_file_ending(path, suffix)
    if not Path(path_with_suffix).exists():
        if suffix == 'yaml':
            # Try loading .yml instead
            return load_yaml(path, 'yml')
        else:
            raise FileNotFoundError(f"Could not load YAML file {path_with_suffix} (Neither with .yaml suffix). "
                                    f"Is the path correct?")

    with open(path_with_suffix, 'r') as f:
        return yaml.load(f, Loader=yaml.FullLoader)


# =========================================================================
# Images
# =========================================================================

def save_img(img: np.ndarray, path: PathType):
    ensure_directory_exists_for_file(path)
    Image.fromarray(img).save(path)


def load_img(path: PathType) -> np.ndarray:
    img = Image.open(path)
    return np.asarray(img)



from time import time


class Timing:
    """
    Usage:

    with Timing() as t:
        ...
    print(f"took {t[0]:0.2} seconds")

    Within the with-statement we can take several measurements by calling t.measure().
    Every measurement will be available after the closure in the list t.
    """

    class TimeMeasurer:

        def __init__(self, start):
            self.start = start
            self.times = []

        def measure(self) -> float:
            now = time()
            measured_time = now - self.start
            self.times.append(measured_time)
            self.start = now
            return measured_time

        def __len__(self) -> int:
            return len(self.times)

        def __getitem__(self, i) -> float:
            return self.times[i]

        def __str__(self) -> str:
            if len(self.times) == 1:
                return f"{self.times[0]:.2f}"
            return f"{self.times}"

    def __enter__(self) -> TimeMeasurer:
        self.time_measurer = Timing.TimeMeasurer(time())
        return self.time_measurer

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.time_measurer.measure()



from typing import List, Union


class Version(object):
    """
    Models a multi-level version similar to SemVer. A version is comprised of numbers separated by dots with an optional
    'v' letter in front.
    Examples: v1.0, 0.0.1, 1.22.333.4444

    The numbers between the dots are referred to as 'levels'. I.e., for v1.0 get_level(0) == 1 and get_level(1) == 0.
    Versions can be compared to their string representations and sorted.
    """

    _levels: List[int]

    def __init__(self, *version_str_or_ints: Union[str, int]):
        if len(version_str_or_ints) == 1 and isinstance(version_str_or_ints[0], str):
            self._levels = Version.parse(version_str_or_ints[0])
        elif len(version_str_or_ints) >= 1 and all([isinstance(part, int) for part in version_str_or_ints]):
            self._levels = version_str_or_ints
        else:
            raise ValueError(f"Version specifier has to a single string or several ints. Got {version_str_or_ints}")

    @staticmethod
    def from_zero(n_levels: int):
        """
        Creates a initial "0" version with the specified amount of levels.
        E.g., 0.0.0

        Parameters
        ----------
            n_levels: how many levels the "0" version should have
        """

        return Version(".".join(["0" for _ in range(n_levels)]))

    @staticmethod
    def from_one(n_levels: int, bump_level: int = 0):
        """
        Creates an initial "1" version with the specified amount of levels.
        E.g., 1.0 or 0.0.1

        Parameters
        ----------
            n_levels: the number of levels
            bump_level: where the 1 should be
        """

        bump_level = bump_level % n_levels
        levels = ["1" if level == bump_level else "0" for level in range(n_levels)]
        return Version(".".join(levels))

    def get_n_levels(self) -> int:
        return len(self._levels)

    def get_level(self, level: int) -> int:
        assert -self.get_n_levels() <= level < self.get_n_levels(), \
            f"Level of bounds. Got `{level}` but only have {self.get_n_levels()} levels"

        return self._levels[level]

    def get_levels(self) -> List[int]:
        return list(self._levels)

    def bump(self, level: int = 0):
        """
        Increases one level by 1 and reverts all lower levels to 0.
        levels are 0-indexed.

        Examples:
        ---------
        >>> version = Version("1.2.3")
        >>> version.bump(0) -> 2.0.0
        >>> version.bump(1) -> 1.3.0
        >>> version.bump(-1) -> 1.2.4

        Parameters
        ----------
            level: which level should be increased
        """

        version_at_level = self.get_level(level)
        level = level % self.get_n_levels()

        self._levels = [v
                        if l < level
                        else
                        (
                            version_at_level + 1
                            if l == level
                            else 0
                        )

                        for l, v in enumerate(self._levels)]

    # -------------------------------------------------------------------------
    # Comparison utilities
    # -------------------------------------------------------------------------

    def __str__(self) -> str:
        return ".".join((str(level) for level in self._levels))

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Version):
            return self._levels == other._levels
        elif isinstance(other, str):
            return str(self) == other
        else:
            return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str):
            other = Version(other)

        assert self.get_n_levels() == other.get_n_levels(), \
            f"Versions differ in number of levels. Got {self.get_n_levels()}" and {other.get_n_levels()}

        # Check version levels step by step
        for level in range(self.get_n_levels()):
            lvl_1 = self.get_level(level)
            lvl_2 = other.get_level(level)
            if lvl_1 < lvl_2:
                return True
            elif lvl_1 > lvl_2:
                return False

        # All levels are equal. This version is not smaller than the other
        return False

    # -------------------------------------------------------------------------
    # Parsing utilities
    # -------------------------------------------------------------------------

    @staticmethod
    def parse(version: str) -> List[int]:
        if version.startswith('v'):
            version = version[1:]
        try:
            levels = [Version._parse_level(level) for level in version.split('.')]
            return levels
        except ValueError:
            raise ValueError(f"`{version}` is not a valid version specifier. "
                             f"Only strings that consists of non-negative numbers separated by points are accepted")

    @staticmethod
    def is_valid(version: str) -> bool:
        try:
            Version.parse(version)
            return True
        except ValueError:
            return False

    @staticmethod
    def _parse_level(level: str) -> int:
        level = int(level)
        if level < 0:
            raise ValueError()
        return level



from abc import ABC, abstractmethod
from typing import Iterator, List

import numpy as np


class ChoiceSampler(ABC, Iterator[int]):

    def __init__(self, n_choices: int):
        self._choices = list(range(n_choices))

    def choice_exhausted(self, choice_idx: int):
        self._choices.remove(choice_idx)

    def get_remaining_choices(self) -> List[int]:
        return self._choices


class SamplingStrategy(ABC):

    @abstractmethod
    def create_sampler(self, n_choices: int) -> ChoiceSampler:
        pass


class RandomSamplingStrategy(SamplingStrategy):
    class RandomChoiceSampler(ChoiceSampler):

        def __next__(self) -> int:
            return np.random.choice(self._choices)

    def create_sampler(self, n_choices: int) -> ChoiceSampler:
        return self.RandomChoiceSampler(n_choices)


class WeightedSamplingStrategy(SamplingStrategy):
    class WeightedSampler(ChoiceSampler):

        def __init__(self, n_choices: int, weights: List[float]):
            assert n_choices == len(weights), \
                f"WeightedSamplingStrategy got {len(weights)} weights, but {n_choices} choices"
            super(WeightedSamplingStrategy.WeightedSampler, self).__init__(n_choices)
            weights = np.array(weights)

            # Completely ignore choices that were given 0 probability
            zero_prob_choices = weights == 0
            weights = weights[~zero_prob_choices]
            self._choices = [choice
                             for choice, zero_prob_choice in zip(self._choices, zero_prob_choices)
                             if not zero_prob_choice]

            self._original_weights = weights / sum(weights)
            self._weights = self._original_weights

        def __next__(self) -> int:
            return np.random.choice(self._choices, p=self._weights)

        def choice_exhausted(self, choice_idx: int):
            super(WeightedSamplingStrategy.WeightedSampler, self).choice_exhausted(choice_idx)

            if len(self.get_remaining_choices()) == 0:
                return

            remaining_weights = self._original_weights[self._choices]
            remaining_weights /= sum(remaining_weights)
            assert sum(remaining_weights) > 0, f"remaining weights (initial: {self._original_weights}) sum to 0"

            self._weights = remaining_weights

    def __init__(self, weights: List[float]):
        self._weights = weights

    def create_sampler(self, n_choices: int) -> ChoiceSampler:
        return self.WeightedSampler(n_choices, self._weights)


class CyclicSamplingStrategy(SamplingStrategy):
    class CyclicSampler(ChoiceSampler):

        def __init__(self, n_choices: int, sample_sequence: List[int]):
            assert min(sample_sequence) >= 0 and max(sample_sequence) < n_choices, \
                f"sample sequence ({sample_sequence}) contains choices that are out of range for {n_choices} choices"
            super().__init__(n_choices)
            self._sample_sequence = list(sample_sequence)
            self._current_idx = 0

        def __next__(self) -> int:
            choice = self._sample_sequence[self._current_idx]
            self._current_idx = (self._current_idx + 1) % len(self._sample_sequence)
            return choice

        def choice_exhausted(self, choice_idx: int):
            super(CyclicSamplingStrategy.CyclicSampler, self).choice_exhausted(choice_idx)

            if len(self.get_remaining_choices()) == 0:
                return

            # 1, 2, 2, 1, 3
            #       x
            # remove 2
            # 1, 1, 3
            #    x

            # Remove all occurrences of the choice from the sample sequence
            filtered_sample_sequence = list()
            for idx, choice in enumerate(self._sample_sequence):
                if idx < self._current_idx:
                    # Ensure that pointer is moved in such a way that sampling proceeds as if choice_idx didn't exist
                    self._current_idx -= 1
                if choice != choice_idx:
                    filtered_sample_sequence.append(choice)

            # Can happen that all choices after current_idx were deleted. Then pointer needs to get back to beginning
            self._current_idx = self._current_idx % len(filtered_sample_sequence)

            self._sample_sequence = filtered_sample_sequence

    def __init__(self, sample_sequence: List[int]):
        self._sample_sequence = sample_sequence

    def create_sampler(self, n_choices: int) -> ChoiceSampler:
        return self.CyclicSampler(n_choices, self._sample_sequence)


class SequentialSamplingStrategy(SamplingStrategy):
    class SequentialSampler(ChoiceSampler):
        def __next__(self) -> int:
            return self._choices[0]

    def create_sampler(self, n_choices: int) -> ChoiceSampler:
        return self.SequentialSampler(n_choices)


from pathlib import Path
from shutil import rmtree
from typing import List, Optional, Union, Generic, TypeVar, Type

from silberstral import reveal_type_var

from elias.folder.folder import Folder
from elias.manager.run import RunManager
from elias.util import ensure_directory_exists

_RunManagerType = TypeVar('_RunManagerType', bound=RunManager)


class RunFolder(Generic[_RunManagerType]):
    """
    Manages several runs stored in a folder. A run can be the result of any arbitrary script.
    Typically the artifacts of several runs of the same script would be stored in the same run folder.
    Each run is identified by an incrementing run id.
    Runs in the same run folder share a name format that can contain the following wildcards:
        - $: for the run id
        - *: for extra name information for a run (optional)
    """

    _name_format: str
    _folder: Folder

    def __init__(self, location: str, name_format: str, localize_via_run_name: bool = False):
        ensure_directory_exists(location)
        self._folder = Folder(location)
        self._name_format = name_format
        self._localize_via_run_name = localize_via_run_name

        self._cls_run_manager: Type[_RunManagerType] = reveal_type_var(self, _RunManagerType)

    def cd(self, sub_folder: str):
        # Inplace cd to avoid having to infer the correct subclass of the current object
        self._folder.cd(sub_folder, inplace=True)

    def list_runs(self) -> List[str]:
        return self._folder.list_file_numbering(self._name_format, return_only_file_names=True)

    def list_run_ids(self) -> List[int]:
        return self._folder.list_file_numbering(self._name_format, return_only_numbering=True)

    def generate_run_name(self, name: Optional[str] = None) -> str:
        return self._folder.generate_next_name(self._name_format, name=name)

    def open_run(self, run_name_or_id: Union[str, int]) -> _RunManagerType:
        run_name = self.resolve_run_name(run_name_or_id)
        return self._cls_run_manager.from_location(self._folder.get_location(),
                                                   run_name,
                                                   localize_via_run_name=self._localize_via_run_name)

    def new_run(self, name: Optional[str] = None) -> _RunManagerType:
        new_run_name = self.generate_run_name(name)
        return self.open_run(new_run_name)

    def delete_run(self, run_name_or_id: Union[str, int]):
        run_name = self.resolve_run_name(run_name_or_id)
        run_folder = f"{self._folder.get_location()}/{run_name}"
        assert Path(run_folder).exists(), f"Cannot delete run {run_name}. It does not exist"
        assert Path(run_folder).is_dir(), f"{run_folder} is not a folder"

        rmtree(run_folder)

    def get_run_name_by_id(self, run_id: int) -> Optional[str]:
        return self._folder.get_file_name_by_numbering(self._name_format, run_id)

    def resolve_run_name(self, run_name_or_id: Union[str, int]) -> Optional[str]:
        if isinstance(run_name_or_id, int):
            return self.get_run_name_by_id(run_name_or_id)
        elif self._folder.file_exists(run_name_or_id):
            # Run name was given in its entirety
            return run_name_or_id
        else:
            # Maybe run name was given without optional parts, e.g., NET-123 instead of NET-123-more-dropout
            # Try to extract run id and then find full file name from that
            file_numbering = self._folder.get_numbering_by_file_name(self._name_format, run_name_or_id)
            if file_numbering is None:
                # Given run_name does not match name format
                return None

            file_name = self._folder.get_file_name_by_numbering(self._name_format, file_numbering)
            return file_name



from pathlib import Path
from typing import Any, List, Dict, Tuple, Type

import matplotlib.pyplot as plt

from elias.util.fs import ensure_file_ending
from elias.util.io import save_pickled, load_pickled


class Analysis:
    _location: str

    def __init__(self, location: str, analysis_name: str):
        analysis_location = f"{location}/{analysis_name}"
        assert Path(analysis_location).is_dir(), \
            f"Could not find directory '{location}'. Is the path correct?"

        self._location = analysis_location

    @classmethod
    def from_location(cls: Type['Analysis'],
                      location: str,
                      analysis_name: str,
                      localize_via_analysis_name: bool = False) -> 'Analysis':
        """
        Creates an analysis for the specified location with default parameters.
        If the subclass constructor takes different arguments than the location, this needs to be overridden
        by subclasses to ensure that there is a instantiation method that takes exactly one argument.

        Parameters
        ----------
            location: path to the folder containing analyses
            analysis_name: name of analysis linking to folder containing analysis artifacts
            localize_via_analysis_name: whether only the analysis name should be used to find the folder

        Returns
        -------
            an analysis of the sub class at the specified location
        """

        try:
            if localize_via_analysis_name:
                new_analysis = cls(analysis_name)
            else:
                new_analysis = cls(location, analysis_name)
        except TypeError:
            raise NotImplementedError(f"Could not construct analysis {cls} with a single location parameter. "
                                      f"Please override from_location() to match the class __init__() method")
        return new_analysis

    def save_pyplot_fig(self, fig_name: str, file_ending: str = 'pdf'):
        plt.savefig(f"{self._location}/{ensure_file_ending(fig_name, file_ending)}")

    def save_object(self, obj: Any, name: str):
        save_pickled(obj, f"{self._location}/{name}")

    def save_objects(self, objects: Dict[str, Any]):
        for name, obj in objects.items():
            self.save_object(obj, name)

    def load_object(self, name: str) -> Any:
        return load_pickled(f"{self._location}/{name}")

    def load_objects(self, *names) -> Tuple[Any]:
        return tuple(self.load_object(name) for name in names)

    def ls(self) -> List[str]:
        return [p.name for p in Path(self._location).iterdir()]





import os
from pathlib import Path


def ensure_file_ending(file: str, ending: str) -> str:
    """
    Ensures that the given file path has the specified ending.

    Parameters
    ----------
        file: path to file
        ending: extension that the file should have

    Returns
    -------
        the file path with the specified extension if it did not already have that ending
    """

    return file if file.endswith(ending) else f"{file}.{ending}"


def ensure_directory_exists_for_file(path: str):
    """
    Ensures that the folder to the specified path exists and creates a nested folder structure if necessary.
    Be careful to use trailing '/' if `path` directly constitutes the target folder (and not an arbitrary file
    within that folder).

    Parameters
    ----------
        path: path to the file or folder for which an underlying directory structure will be ensured
    """

    Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)


def ensure_directory_exists(path: str):
    """
    Ensures that the specified folder exists and creates a nested folder structure if necessary.

    Parameters
    ----------
        path: path to the folder which should exist
    """

    Path(path).mkdir(parents=True, exist_ok=True)



import re
from os import mkdir
from pathlib import Path
from shutil import rmtree
from typing import List, Union, Tuple, Optional


# TODO: Allow having leading zeros for $

class Folder:
    _location: str

    def __init__(self, location: str):
        assert Path(f"{location}").is_dir(), \
            f"Could not find directory '{location}'. Is the location correct?"

        self._location = location

    def cd(self, sub_folder: str, inplace: bool = False) -> 'Folder':
        """
        Switches to the specified sub folder.

        Parameters
        ----------
            sub_folder:
                where to cd to
            inplace:
                whether the current folder object should be altered or a new one should be created

        Returns
        -------
            A new folder object at the specified location
        """

        resolved_sub_folder_path = str(Path(f"{self._location}/{sub_folder}").resolve())
        if inplace:
            self._location = resolved_sub_folder_path
            return self
        else:
            return self.__init__(resolved_sub_folder_path)

    def ls(self, name_format: Optional[str] = None) -> List[str]:
        if name_format is None:
            return [p.name for p in Path(self._location).iterdir()]
        else:
            return self.list_file_numbering(name_format, return_only_file_names=True)

    def mkdir(self, folder_name: str):
        mkdir(f"{self._location}/{folder_name}")

    def rmdir(self, folder_name: str):
        rmtree(f"{self._location}/{folder_name}")

    def get_location(self) -> str:
        return self._location

    def file_exists(self, file_name: str) -> bool:
        return Path(f"{self._location}/{file_name}").exists()

    def list_file_numbering(self,
                            name_format: str,
                            return_only_numbering: bool = False,
                            return_only_file_names: bool = False) -> List[Union[Tuple[int, str], int, str]]:
        r"""
        Finds all (numbered) files/folder in the specified directory that match the name_format and returns them in
        sorted fashion according to the number in the file/folder name.
        The numbering of the file/folder will also be returned.
        The passed name_format is expected to have exactly one occurrence of `$` which controls where in the file/folder
        name the number should appear. Additionally, `*` is supported as a wildcard similar to glob.
        This method solves the problem of numbers in file/folder names not being treated as "numbers" by the OS but
        rather as strings, i.e., 2-apple will appear after 10-banana although 2 < 10.

        Examples
        --------
            Given Folder structure:

            root
             ├── analysis-batch-norm-100-lambda-10
             ├── analysis-batch-norm-50-lambda-9
             ├── epoch-11.ckpt
             ├── epoch--1.ckpt
             ├── P2P-10
             └── P2P-9

             >>> folder = Folder('root')
             >>> folder.list_file_numbering('P2P-$', return_only_file_names=True)
             >>> >>> ['P2P-9', 'P2P-10']

             >>> folder.list_file_numbering('epoch-$.ckpt')
             >>> >>> [(-1, 'epoch--1.ckpt'), (11, 'epoch-11.ckpt')]

             >>> folder.list_file_numbering('analysis-*-$')
             >>> >>> [(9, 'analysis-batch-norm-50-lambda-9'), (10, 'analysis-batch-norm-100-lambda-10')]

        Parameters
        ----------
            name_format:
                specifies which file/folder names should be filtered and where in their name the numbering occurs.
                Two types of format specifiers are supported:
                    $: indicates where in the file name the number is written
                    *: matches arbitrary text
            return_only_numbering:
                whether to only return the numbering of the matched files
            return_only_file_names:
                whether to only return the name of the matched files

        Returns
        -------
            A sorted list of (numbering, file name) pairs for each matching file/folder in the passed directory
        """

        assert not (return_only_numbering and return_only_file_names), \
            "Can only set one of return_only_numbering and return_only_file_names"

        regex = self._build_numbering_extraction_regex(name_format)
        file_names = [file.name for file in Path(self._location).iterdir()]
        file_names_and_numbering = [(int(regex.search(file_name).group(1)), file_name)
                                    for file_name in file_names if regex.match(file_name)]
        file_names_and_numbering = sorted(file_names_and_numbering, key=lambda x: x[0])

        if return_only_numbering:
            return [file_name_and_numbering[0] for file_name_and_numbering in file_names_and_numbering]
        elif return_only_file_names:
            return [file_name_and_numbering[1] for file_name_and_numbering in file_names_and_numbering]
        else:
            return file_names_and_numbering

    def get_file_name_by_numbering(self, name_format: str, numbering: int) -> Optional[str]:
        """
        Obtains the corresponding file name given its numbering.

        Parameters
        ----------
            name_format:
                The assumed format of ordered files. See list_file_numbering()
            numbering:
                the numbering of the desired file

        Returns
        -------
            - The corresponding file name of the given numbering if it could be found
            - None, otherwise
        """

        file_numberings_and_names = self.list_file_numbering(name_format)
        for file_numbering, name in file_numberings_and_names:
            if file_numbering == numbering:
                return name

        # File numbering could not be found
        return None

    def get_numbering_by_file_name(self, name_format: str, file_name: str) -> Optional[int]:
        """
        Extracts the numbering of a file name. The file does not have to exist.

        Parameters
        ----------
            name_format:
                The assumed format of ordered files. See list_file_numbering()
            file_name:
                the name of the desired file for which the numbering should be extracted

        Returns
        -------
            - The corresponding numbering of the file if it could be extracted
            - None, otherwise
        """

        regex = self._build_numbering_extraction_regex(name_format)
        if regex.match(file_name):
            return int(regex.search(file_name).group(1))
        else:
            return None

    @staticmethod
    def substitute(name_format: str, numbering: int, name: Optional[str] = None) -> str:
        """
        Substitutes numbering and optional free text names into the given `name_format`.
        Replaces $ with numbering and potential * with name.

        Parameters
        ----------
            name_format:
                format used as a basis for substitution
            numbering:
                numbering to substitute in
            name:
                optional free text name to substitute in

        Returns
        -------
            The name format with any $ and * wildcards replaced by numbering and name
        """

        wildcard_present = '*' in name_format
        name_none = name is None
        assert wildcard_present ^ name_none, \
            "If `name` is given, `*` should appear in `name_format` and vice-versa"
        assert name_format.count('*') <= 1, 'Wildcard `*` cannot appear more than once'
        substituted_name = name_format.replace('$', f'{numbering}')

        if wildcard_present:
            substituted_name = substituted_name.replace('*', name)

        return substituted_name

    def generate_next_name(self,
                           name_format: str,
                           name: Optional[str] = None,
                           create_folder: bool = True) -> str:
        """
        Generates a new run name (and per default creates a folder with that name) that follows the given `name_format`
        and has a numbering that is one larger than the highest present in the folder.
        If this will be the first folder matching the given `name_format` its numbering will be 1.
        The `name_format` assumes exactly one `$` symbol for the location of the numbering in the name. Optionally,
        a single `*` can be specified that will be filled with the given `name`.

        Examples
        --------

            Given Folder structure:

            root
             ├── analysis-batch-norm-100-lambda-10
             ├── analysis-batch-norm-50-lambda-9
             ├── epoch-11.ckpt
             ├── epoch--1.ckpt
             ├── P2P-10
             └── P2P-9

            >>> folder = Folder('root')
            >>> folder.generate_next_name('P2P-$')
            >>> >>> 'P2P-11'

            >>> folder.generate_next_name('epoch-$.ckpt', create_folder=False)
            >>> >>> 'epoch-12.ckpt'

            >>> folder.generate_next_name('analysis-*-$', name='batch-norm-no-lambda')
            >>> >>> 'analysis-batch-norm-no-lambda-11'

        Parameters
        ----------
            name_format:
                Specifies which folders/files in the directory will be considered for the numbering and also formats
                the newly generated run name.
                    $: location of numbering
                    * (optional, max 1): wildcard text. If present, `name` should be set
            name:
                In the case, a `*` wildcard is present in the `name_format` a name has to be specified for the newly
                generated run
            create_folder:
                whether a folder with the newly generated run name should be created automatically

        Returns
        -------
            The name of the new run, ensuring an ascending numbering
        """

        file_numbering = self.list_file_numbering(name_format, return_only_numbering=True)
        if len(file_numbering) == 0:
            new_id = 1
        else:
            max_id = max(file_numbering)
            new_id = max_id + 1 if max_id > 0 else 1  # If only negative IDs are present, use 1

        new_name = self.substitute(name_format, new_id, name=name)

        if create_folder:
            self.mkdir(new_name)

        return new_name

    @staticmethod
    def _build_numbering_extraction_regex(name_format: str) -> re.Pattern:
        assert name_format.count('$') == 1, "The number specifier '$' has to appear in the passed format exactly once"
        assert name_format.count('[') == name_format.count(']'), "square brackets not matching in name format"

        name_format = re.escape(name_format)

        # \[...\] -> (...)?
        name_format = name_format.replace('\\[', '(')
        name_format = name_format.replace('\\]', ')?')

        # $ -> Numbering format
        name_format = name_format.replace(r'\$', r'(-?\d+)')

        # * -> Wildcard format. Make * non-greedy with ?.
        # Otherwise it would match trailing minus signs in '*-$.ckpt' such that $ could never be a negative number
        name_format = name_format.replace(r'\*', r'.*?')

        # Ensure that name_format matches exactly without any leading/trailing leftovers
        name_format = f'^{name_format}$'

        regex = re.compile(name_format)
        return regex



from typing import Iterable, List, Optional, Iterator, TypeVar

import numpy as np

from elias.data.loader import RandomAccessDataLoader
from elias.data.sampling import SamplingStrategy, CyclicSamplingStrategy, RandomSamplingStrategy, \
    WeightedSamplingStrategy, SequentialSamplingStrategy, ChoiceSampler
from elias.data.stop_criterion import CombinedIterableStopCriterion, CombinedIterableStopCriterionAllEmpty

_T = TypeVar('_T')


class CombinedIterableDataLoader(Iterable[_T]):
    def __init__(self,
                 data_loaders: List[Iterable[_T]],
                 shuffle: bool = False,
                 sample_weights: Optional[List[float]] = None,
                 stop_criterion: CombinedIterableStopCriterion = CombinedIterableStopCriterionAllEmpty(),
                 alternating_sampling: bool = False,
                 sampling_strategy: Optional[SamplingStrategy] = None,
                 return_dl_idx: bool = True):
        """
        Combines the specified iterables into a single iterable dataloader. Per default, the given iterables will
        be traversed in order. If :paramref:`~params.shuffle` is set, the dataloaders will be traversed randomly.

        Parameters
        ----------
            data_loaders:
                the iterables that should be combined
            shuffle:
                whether the iterables will be traversed in order or randomly
            sample_weights:
                If :paramref:`~params.shuffle` is `True`, the sample weights specify for each dataloader how likely
                it should be to draw from it. If it is `None` elements will be drawn uniformly from the given iterables
            stop_criterion:
                specifies under what circumstances the combined dataloader should stop. E.g., whether it should exhaust
                all given dataloaders, stop when any of the iterables is empty or continue until a specific dataloader
                is empty
            alternating_sampling:
                Overrides `sample_weights`. If set to `True`, provided iterators will be traversed in a round robin
                style instead of exhausting the first before continuing with the second.
            return_dl_idx:
                whether the returned elements of the combined dataloader should be a tuple containing the index of the
                original dataloader and the actual element (similar to Python's enumerate()).
        """

        self._data_loaders = data_loaders
        self._shuffle = shuffle
        self._return_dl_idx = return_dl_idx
        self._alternating_sampling = alternating_sampling

        if sampling_strategy is not None:
            self._sampling_strategy = sampling_strategy
        elif alternating_sampling:
            # sample from each data loader once before repeating
            self._sampling_strategy = CyclicSamplingStrategy(list(range(len(data_loaders))))
        elif shuffle:
            if sample_weights is None:
                # uniform random sampling
                self._sampling_strategy = RandomSamplingStrategy()
                # self._sample_weights = np.array([1 / len(data_loaders) for _ in range(len(data_loaders))])
            else:
                self._sampling_strategy = WeightedSamplingStrategy(sample_weights)
                # self._sample_weights = None if sample_weights is None else np.array(sample_weights) / sum(
                #     sample_weights)
                assert sample_weights is None or len(sample_weights) == len(data_loaders), \
                    f"Need to specify as many sample weights (got {len(sample_weights)}) " \
                    f"as dataloaders ({len(data_loaders)})"
        else:
            assert sample_weights is None, f"shuffle has to be set, if sample_weights are used"
            # self._sample_weights = None
            self._sampling_strategy = SequentialSamplingStrategy()

        self._stop_criterion = stop_criterion

    def save(self, data, **kwargs):
        raise Exception('CombinedDataManager cannot save')

    def __iter__(self) -> Iterator[_T]:
        return CombinedIterableDataLoader.Iterator([iter(data_manager) for data_manager in self._data_loaders],
                                                   self._sampling_strategy.create_sampler(len(self._data_loaders)),
                                                   self._stop_criterion,
                                                   self._return_dl_idx)

    class Iterator:

        def __init__(self,
                     iterators: List[Iterator[_T]],
                     data_loader_sampler: ChoiceSampler,
                     stop_criterion: CombinedIterableStopCriterion,
                     return_dl_idx: bool):
            self._iterators = iterators
            self._data_loader_sampler = data_loader_sampler
            self._stop_criterion = stop_criterion
            self._return_dl_idx = return_dl_idx
            self._last_chosen_iterator_idx = -1

            # self._identifiers = list(range(len(iterators)))

        def __next__(self) -> _T:
            if len(self._data_loader_sampler.get_remaining_choices()) == 0:
                raise StopIteration()

            iterator_idx = next(self._data_loader_sampler)

            # if self._alternating_sampling:
            #     idx = (self._last_chosen_iterator_idx + 1) % len(self._identifiers)
            #     iterator_idx = self._identifiers[idx]
            #     self._last_chosen_iterator_idx = idx
            # elif self._sample_weights is not None:
            #     sample_weights = self._sample_weights[self._identifiers]
            #     assert sum(sample_weights) > 0, f"sample_weights (initial: {self._sample_weights}) sum to 0"
            #     sample_weights /= sum(sample_weights)
            #     iterator_idx = np.random.choice(self._identifiers, p=sample_weights)
            # else:
            #     # Iterate through all iterators in order
            #     iterator_idx = self._identifiers[0]

            try:
                sample = next(self._iterators[iterator_idx])
                if self._return_dl_idx:
                    return iterator_idx, sample
                else:
                    return sample
            except StopIteration:
                # self._identifiers.remove(iterator_idx)
                # self._last_chosen_iterator_idx -= 1  # Ensure that alternating sampling will not jump over next iterator
                self._data_loader_sampler.choice_exhausted(iterator_idx)
                if self._stop_criterion.should_stop(iterator_idx, self._data_loader_sampler.get_remaining_choices()):
                    raise StopIteration()
                else:
                    return next(self)


class CombinedRandomAccessDataLoader(RandomAccessDataLoader[_T]):

    def __init__(self, data_loaders: List[RandomAccessDataLoader[_T]], shuffle=False):
        # TODO: sample_weights
        self._data_loaders = data_loaders
        self._shuffle = shuffle

        if shuffle:
            self._shuffled_indices = list(range(len(self)))
            np.random.shuffle(self._shuffled_indices)

    def __iter__(self) -> Iterator[_T]:
        return (self[idx] for idx in range(len(self)))

    def _get_single_item(self, idx: int) -> _T:
        assert -len(self) <= idx < len(
            self), f"Index {idx} is out of bounds for combined data loader of size {len(self)}"
        if self._shuffle:
            idx = self._shuffled_indices[idx]

        dl_idx, sample_idx = self._get_dl_idx_for_sample(idx)
        sample = self._data_loaders[dl_idx][sample_idx]

        return dl_idx, sample

    def __len__(self) -> int:
        return sum([len(data_loader) for data_loader in self._data_loaders])

    def _get_dl_idx_for_sample(self, idx: int):
        assert -len(self) <= idx < len(
            self), f"Index {idx} is out of bounds for combined data loader of size {len(self)}"
        seen_samples = 0
        current_dl_idx = 0
        if idx < 0:
            idx += len(self)
        for data_loader in self._data_loaders:
            seen_samples += len(data_loader)
            if idx < seen_samples:
                return current_dl_idx, idx - (seen_samples - len(data_loader))
            current_dl_idx += 1


from .fs import ensure_file_ending, ensure_directory_exists_for_file, ensure_directory_exists
from .io import load_pickled, save_pickled, load_yaml, save_yaml, load_json, save_json, load_zipped_json, \
    save_zipped_json, load_zipped_object, save_zipped_object
from .timing import Timing
from .version import Version



from pathlib import Path
from typing import Type


class RunManager:

    def __init__(self, location: str, run_name: str):

        run_location = f"{location}/{run_name}"
        assert Path(run_location).is_dir(), \
            f"Could not find directory '{location}'. Is the path correct?"

        self._location = run_location
        self._run_name = run_name

    @classmethod
    def from_location(cls: Type['RunManager'],
                      location: str,
                      run_name: str,
                      localize_via_run_name: bool = False) -> 'RunManager':
        """
        Creates a run manager for the specified location with default parameters.
        If the subclass constructor takes different arguments than the location, this needs to be overridden
        by subclasses to ensure that there is a instantiation method that takes exactly one argument.

        Parameters
        ----------
            location: path to the folder containing runs
            run_name: name of run that links to folder containing artifacts
            localize_via_run_name: whether only the run name is used to localize the folder

        Returns
        -------
            a run manager of the sub class at the specified location
        """

        try:
            if localize_via_run_name:
                run_manager = cls(run_name)
            else:
                run_manager = cls(location, run_name)
        except TypeError:
            raise NotImplementedError(f"Could not construct run manager {cls} with a single location parameter. "
                                      f"Please override from_location() to match the class __init__() method")
        return run_manager

    def get_run_name(self) -> str:
        return self._run_name



import random
from abc import abstractmethod, ABC
from math import ceil
from typing import Iterable, TypeVar, Generic, List, Generator, Iterator, Type, Union, Any

from silberstral import reveal_type_var

from elias.manager.artifact import ArtifactManager, ArtifactType
from elias.config import Config
from elias.data.loader import RandomAccessDataLoader
from elias.folder.folder import Folder

_ConfigType = TypeVar('_ConfigType', bound=Config)
_StatisticsType = TypeVar('_StatisticsType', bound=Config)
_SampleType = TypeVar('_SampleType')
_T = TypeVar('_T')


# TODO: We can abstract the filesystem requirement away by having an interface that provides iterating over the data
#  and querying for a list of file/data/sample/image names

class BaseDataManager(Generic[_SampleType, _ConfigType, _StatisticsType], ArtifactManager, Iterable[_SampleType], ABC):
    """
    A DataManager provides access to dataset files stored on a file system.
    It is assumed that all dataset files reside in the same root folder.
    Optionally, the DataManager can process additional information of the dataset stored in JSON files:
        - ``config.json``: Contains all information necessary to reproduce the creation of the dataset
        - ``stats.json``: Contains statistics of the dataset that can only be obtained by a potentially costly iteration over the full dataset

    Type Vars
    ---------
        _SampleType:
            The type of samples being returned by this data manager
        _ConfigType:
            The class of the dataset configuration (stored in ``config.json``) which is assumed to be a dataclass
            subclassing :class:`elias.config.Config`. :meth:`save_config` and :meth:`load_config` take/retrieve the
            dataset configuration as a Python object of this class.
        _StatisticsType:
            Same explanation as for `_ConfigType` just for the dataset statistics ``stats.json``
    """

    _data_folder: Folder
    _file_name_format: str
    _shuffle: bool
    _config_cls: _ConfigType
    _statistics_cls: _StatisticsType

    def __init__(self,
                 data_location: str,
                 run_name: str,
                 file_name_format: str,
                 shuffle: bool = False,
                 artifact_type: ArtifactType = ArtifactType.JSON):
        """
        Parameters
        ----------
            data_location: str
                Root folder of stored data. This is where ``config.json`` and ``stats.json`` will be loaded from.
            run_name: str
                the run name or version of the stored data
            shuffle: bool, default False
                Only relevant when using :meth:`_lazy_load_slices`. If shuffle is set to ``True`` dataset files will
                be loaded in random order
            file_name_format:
                Format of the files in the dataset folder. As these are typically numbered, specifying a
                file name format allows convenient loading and saving of dataset files.
                An example format may be: image_$.png, sample-$.txt or dataset-$.p
        """
        super(BaseDataManager, self).__init__(f"{data_location}/{run_name}", artifact_type=artifact_type)

        self._data_folder = Folder(f"{data_location}/{run_name}")
        self._root_location = data_location
        self._run_name = run_name
        self._file_name_format = file_name_format
        self._shuffle = shuffle
        self._config_cls = reveal_type_var(self, _ConfigType)
        self._statistics_cls = reveal_type_var(self, _StatisticsType)

    @classmethod
    def from_location(cls: Type['BaseDataManager'],
                      location: str,
                      dataset_name: str,
                      localize_via_run_name: bool = False) -> 'BaseDataManager':
        """
        Creates a data manager for the specified location with default parameters.
        Needs to be overridden by subclasses.

        Parameters
        ----------
            location:
                path to the folder containing dataset versions
            dataset_name:
                name of the dataset linking to the folder containing preprocessed data, statistics and config files
            localize_via_run_name:
                whether only the dataset name should be used to find the folder

        Returns
        -------
            a data manager of the sub class at the specified location
        """

        try:
            if localize_via_run_name:
                data_manager = cls(dataset_name)
            else:
                data_manager = cls(location, dataset_name)
        except TypeError:
            raise NotImplementedError(f"Could not construct data manager {cls} with a single location parameter. "
                                      f"Please override from_location() to match the class __init__() method")
        return data_manager

    @staticmethod
    def to_batches(generator: Iterable[_T], batch_size: int, lazy: bool = False) -> Generator[List[_T], None, None]:
        """
        Lazyly evaluated batch-wise loading
        """

        if batch_size == 1:
            for item in generator:
                yield item
            return

        if lazy:
            # Lazy returns batches as a generator where objects are only touched upon actually querying them
            iterator = iter(generator)
            try:
                while True:
                    first = next(iterator)

                    def chunk():
                        try:
                            yield first
                            for _ in range(batch_size - 1):
                                yield next(iterator)
                        except StopIteration:
                            pass

                    yield chunk()
            except StopIteration:
                pass
        else:
            # Regular mode materializes all objects within a batch before the batch is returned as a list
            batch = []
            for i, item in enumerate(generator):
                batch.append(item)
                if (i + 1) % batch_size == 0:
                    yield batch
                    batch = []
            if batch:
                yield batch

    @staticmethod
    def batchify_tensor(tensor, batch_size: int) -> Iterator:
        try:
            n_samples = len(tensor)
        except Exception:
            try:
                n_samples = tensor.shape[0]
            except Exception:
                raise ValueError(f"Cannot infer length of passed tensor with type {type(tensor)}. "
                                 f"Ensure to use a common Tensor/Array format")

        n_batches = ceil(n_samples / batch_size)
        for i_batch in range(n_batches):
            if i_batch == n_batches - 1:
                yield tensor[i_batch * batch_size:]  # Return all remaining samples as the last batch
            else:
                yield tensor[i_batch * batch_size: (i_batch + 1) * batch_size]

    def iter_batched(self, batch_size: int) -> Iterator[_SampleType]:
        return self.to_batches(self, batch_size)

    def load_config(self) -> _ConfigType:
        json_config = self._load_artifact("config")
        return self._config_cls.from_json(json_config)

    def save_config(self, config: _ConfigType):
        self._save_artifact(config.to_json(), "config")

    def load_stats(self) -> _StatisticsType:
        json_statistics = self._load_artifact("stats")
        return self._statistics_cls.from_json(json_statistics)

    def save_stats(self, stats: _StatisticsType):
        self._save_artifact(stats.to_json(), "stats")

    def get_file_name_by_id(self, file_id: int) -> str:
        return self._data_folder.substitute(self._file_name_format, file_id)

    def get_location(self) -> str:
        return self._data_folder.get_location()

    @abstractmethod
    def __iter__(self) -> Iterator[_SampleType]:
        pass

    @abstractmethod
    def _save(self, data: Any):
        pass


class BaseSampleDataManager(BaseDataManager[_SampleType, _ConfigType, _StatisticsType]):
    """
    Assumes that all samples lie individually in the data folder.
    """

    def save_sample(self, data: _SampleType, **kwargs):
        next_file_name = self._data_folder.generate_next_name(self._file_name_format, create_folder=False)
        self._save_sample(data, f"{self._data_folder.get_location()}/{next_file_name}")

    def load_sample(self, file_name_or_id: Union[str, int]) -> _SampleType:
        if isinstance(file_name_or_id, int):
            file_name = self._data_folder.get_file_name_by_numbering(self._file_name_format, file_name_or_id)
        else:
            file_name = file_name_or_id

        return self._load_sample(f"{self._data_folder.get_location()}/{file_name}")

    def __iter__(self) -> Iterator[_SampleType]:
        file_names = self._data_folder.list_file_numbering(self._file_name_format, return_only_file_names=True)

        if self._shuffle:
            random.shuffle(file_names)
        if not file_names:
            raise Exception(f"No dataset files found in {self._data_folder.get_location()}. Is the path correct?")

        # Converts the path list into a generator
        for file_name in file_names:
            yield self._load_sample(f"{self._data_folder.get_location()}/{file_name}")

    @abstractmethod
    def _save_sample(self, data: _SampleType, file_path: str):
        pass

    @abstractmethod
    def _load_sample(self, file_path: str) -> _SampleType:
        pass

    def _save(self, data: Any):
        self.save_sample(data)


class BaseSliceDataManager(BaseDataManager[_SampleType, _ConfigType, _StatisticsType]):
    """
    Assumes that the dataset is split into so-called "slices" that themselves contain small parts of the dataset.
    For example, the dataset may be split into several pickled files where each contains 500 samples.
    """

    def save_dataset_slice(self, dataset_slice: Iterator[_SampleType]):
        next_slice_name = self._data_folder.generate_next_name(self._file_name_format, create_folder=False)
        self._save_dataset_slice(dataset_slice, f"{self._data_folder.get_location()}/{next_slice_name}")

    def load_dataset_slice(self, slice_name_or_id: Union[str, int]) -> Iterable[_SampleType]:
        if isinstance(slice_name_or_id, int):
            slice_name = self._data_folder.get_file_name_by_numbering(self._file_name_format, slice_name_or_id)
        else:
            slice_name = slice_name_or_id

        return self._load_dataset_slice(f"{self._data_folder.get_location()}/{slice_name}")

    def __iter__(self) -> Iterator[_SampleType]:
        slice_names = self._data_folder.list_file_numbering(self._file_name_format, return_only_file_names=True)

        if self._shuffle:
            random.shuffle(slice_names)
        if not slice_names:
            raise Exception(f"No dataset files found in {self._data_folder.get_location()}. Is the path correct?")

        # Converts the path list into a generator
        for slice_name in slice_names:
            for sample in self._load_dataset_slice(f"{self._data_folder.get_location()}/{slice_name}"):
                yield sample

    @abstractmethod
    def _save_dataset_slice(self, dataset_slice: Iterator[_SampleType], slice_path: str):
        pass

    @abstractmethod
    def _load_dataset_slice(self, slice_name: str) -> Iterable[_SampleType]:
        pass

    def _save(self, data: Any):
        self.save_dataset_slice(data)


class RandomAccessSampleDataManager(RandomAccessDataLoader[_SampleType],
                                    BaseSampleDataManager[_SampleType, _ConfigType, _StatisticsType],
                                    ABC):
    pass


class RandomAccessSliceDataManager(RandomAccessDataLoader[_SampleType],
                                   BaseSliceDataManager[_SampleType, _ConfigType, _StatisticsType],
                                   ABC):
    pass
